
# required module init file
#==============================================================================
#
#  $Id$
#
"""
    Wrapper classes for databases that hide the underlying implementation
    details.
 
    Usage:
 
        from odb import Env
  
        # open or create a database environment
        env = Env('mydb')
  
        # get the B-Tree database "data"
        data = env.getBTreeDB('data')
  
        # store objects in it
        data['first'] = 100
        data['second'] = "this is a string"
        data['third'] = MyClass(1, 2, 3)
        
        # transactional updates
        txn = env.getTxn()
        data.set(txn, 'atomic', 'more')
        data.set(txn, 'commit', 'stuff')
        txn.commit() # or "txn.abort()" to roll both changes back
        
        # cursors
        for key, value in data.cursor():
    
      
"""
#
#  Copyright (C) 2006 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import odb, sys, os

BDB = 0
MEM = 1
FULL = 2

def _pickDatabase(path, restore, prefer = FULL):

    # if the path exists, we're dealing with an existing database
    if os.path.exists(path):
        
        # check for full
        if os.path.exists(os.path.join(path, '.full')):
            import full
            return full.FullEnv(path)
            
        # check for mem
        if os.path.exists(os.path.join(path, '.mem')):
            import mem
            return mem.MemEnv(path)

        # check for berkeley
        for file in os.listdir(path):
            import bdb
            if bdb._logRx.match(file):
                return bdb.BDBEnv(path, restore)

        raise ValueError('Directory %s exists but does not contain a '
                          'recognizable database' % path
                         )

    # create based on preference
    elif prefer == BDB:
        import bdb
        return bdb.BDBEnv(path, restore)

    elif prefer == MEM:
        import mem
        return mem.MemEnv(path)
    
    elif prefer == FULL:
        import full
        return full.FullEnv(path)

    else:
        raise ValueError("bad value for 'prefer': %s" % prefer)

class Env:
    """
        A database environment.  Groups together a set of databases.
    """

    def __init__(self, path, recover = False, prefer = FULL):
        self.__impl = _pickDatabase(path, recover, prefer)

    def getBTreeDB(self, name, allowDups = False):
        return self.__impl.getBTreeDB(name, allowDups)

    def getHashDB(self, name, allowDups = False):
        return self.__impl.getHashDB(name, allowDups)

    def getTxn(self):
        """
	    Returns a new transaction for the environment.  This is a raw BDB
	    transaction.  See the @Txn class for documentation.
        """
        return self.__impl.getTxn()

    def startDeadlockThread(self, deadlockCheckInterval, noticeCallback):
        """
            Start the deadlock detection thread.

            [Added for berkeley db, need to do something different]
        """
        self.__impl.startDeadlockThread(deadlockCheckInterval, noticeCallback)

    def getDatabaseFiles(self):
        """
            Returns the list of the names of all of the files for the 
            database. """
        return self.__impl.getDatabaseFiles()

    def getAllLogs(self):
        """
            Returns a list of the names of all of the logfiles for the 
            database. """
        return self.__impl.getAllLogs()
    
    def checkpoint(self):
        """
            Forces checkpointing of the database.  Checkpointing is when we 
            transfer the in-memory data structures to their backing store 
            files.  If you don't checkpoint occasionally, you will have really 
            long startup times and have to keep the entire database in memory 
            all of the time.
        """
        self.__impl.checkpoint()

# vim:sw=4:
