#==============================================================================
#
#  $Id$
#
"""
   Full implementation: does everything that mem does without having to keep
   the entire environment in memory or write all of it during a checkpoint:
   loads the databases on demand and keeps a configurable subset in memory.
   During checkpointing, writes only the dirty parts.
   
   We need a general organizing philosophy for transactional writes:
   
   1) when anything is created, it has no position in the store
   2) positions must be allocated when the object is written, and before any
      parent node is written
"""
#
#  Copyright (C) 2007 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import os, struct, array
from cStringIO import StringIO
from pickle import Pickler, Unpickler
from dbfile import HeapFile
from mem import MemEnv, MemBTreeDB, BTreeNode, KeyVal
import mem # XXX for _trace

def trace(msg):
   return
   print msg

class Stateful:

   # Node states
   STUB = 0         # the node has not been loaded yet - just a position
   CLEAN = 1        # the node is the same as what is persisted
   DIRTY = 2        # the node needs to be persisted

def _createKeyVal(storage, key, val, state = Stateful.DIRTY, 
                  keyStorage = None):
   if state != Stateful.STUB:
      stream = StringIO()
      Pickler(stream).dump(val)
      data = stream.getvalue()
      storage = storage.alloc(len(data))
   else:
      data = None
   assert keyStorage is not None
   return FullKeyVal(storage, key, val, data, state, keyStorage)

class FileWrapper:
   """
      Just so we can figure out where we're writing from.
   """
   
   debugEnabled = False
   def __init__(self, fp, writable):
      self.__fp = fp
      self.writable = writable
   
   def debug(self, msg):
      if self.debugEnabled: print 'FileWrapper: %s' % msg
      
   def write(self, data):
      if not self.writable:
         raise Exception('Not in write mode')
      self.debug('at %d writing %s' % (self.__fp.tell(), repr(data)))
      self.__fp.write(data)
   
   def read(self, size):
      pos = self.__fp.tell()
      data = self.__fp.read(size)
      self.debug('from %d read %d:%s' % (pos, size, repr(data)))
      return data
   
   def seek(self, pos, whence = 0):
      self.__fp.seek(pos, whence)
      self.debug('seeking %d, at %d' % (pos, self.__fp.tell()))
   
   def tell(self):
      return self.__fp.tell()
   
   def truncate(self, size):
      if not self.writable:
         raise Exception('Not in write mode')
      self.__fp.truncate(size)
   
   def flush(self):
      self.__fp.flush()
   
class _Storage:
   
   class Common:
      def __init__(self, file):
         self.file = file
         self.keyStorageByPos = {}
         self.keyStorageByKey = {}
         self.deleted = []
         self.dirty = set()

   def _handlePosAssigned(self, pos):
      """
         Lets derived clases do their own thing when the position is assigned.
         
         Base class implementation does nothing.
      """
      pass
      
   def __get_pos(self):
      if not self.__pos:
         self.__pos = self.common.file.alloc(self.size)
         self._handlePosAssigned(self.__pos)
      return self.__pos

   pos = property(__get_pos)

   commonsForFile = {}
   def __init__(self, file = None, pos = None, common = None, size = None):
      self.__pos = pos
      self.size = size
      if common:
         self.common = common
      else:
         self.common = self.Common(file)
         if self.commonsForFile.has_key(file):
            raise Exception('duplicate commons')
         self.commonsForFile[file] = self.common
      if pos is not None:
         self._handlePosAssigned(pos)
   
   def isAllocated(self):
      """
         returns true if the storage is allocated (has a position in the 
         underlying file storage mechanism.
      """
      return self.__pos and True or False

   def write(self, data):
      self.common.file.seek(self.pos)
      self.common.file.write(data)
      self.common.file.flush()

   def read(self):
      data = self.common.file.get(self.pos)
      self.size = len(data)
      return data

   def getKey(self, keyPos):
      """
         Returns the key (@_KeyStorage instance) at the given position.
      """
      # try to resolve in the cache, if not found go to the file system
      if self.common.keyStorageByPos.has_key(keyPos):
         return self.common.keyStorageByPos[keyPos]
      else:

         # create a new key
         keyStorage = _KeyStorage(self.common.file, keyPos, self.common)
         return keyStorage
   
   def getKeyStorage(self, key):
      """
         Returns a @_KeyStorage instance for the key.
      """
      storage = self.common.keyStorageByKey.get(key)
      if storage:
         return storage
      else:
         storage = _KeyStorage(self.common.file, common = self.common,
                               key = key,
                               refCount = 1
                               )
         return storage

   def addKey(self, key):
      """
         Adds a new key to the cache and stores it in the storage device.
         Returns the key storage (@_KeyStorage).
      """
      # see if we've already got this key
      storage = self.common.keyStorageByKey.get(key)
      if storage:
         return storage
      else:
         # we don't: store it
         trace('creating storage for key %s' % key)
         storage = _KeyStorage(self.common.file, common = self.common,
                               key = key,
                               refCount = 1
                               )
         return storage

   def alloc(self, size):
      """
         Returns a new storage instance capable of storing a block of the
         given size.  The actual allocation of the block may be deferred until
         the write.
      """
      return _Storage(self.common.file, common = self.common, size = size)

   def create(self, pos):
      """
         Returns a new storage instance derived from this one but with a
         different position.
      """
      return _Storage(common = self.common, pos = pos)
   
   def markDeleted(self):
      """
         Adds the storage node to the deleted list of the commons.
      """
      # we only need to worry about deleting this if it is actually allocated
      if self.__pos:
         trace('%s marking %s for deletion' % (self.common, self.__pos))
         self.common.deleted.append(self)
      elif self in self.common.dirty:
         # the storage is dirty (scheduled to be deleted).  Remove it from the 
         # dirty set.
         trace('removing dirty %s' % self)
         self.common.dirty.remove(self)
   
   def flush(self):
      """
         Flush all changes recorded in the common object.
      """
      # remove all deleted items
      trace('flushing %s' % self.common)
      for storage in self.common.deleted:
         self.common.file.free(storage.pos)
      self.common.deleted = []
      
      # write all dirty items
      for item in self.common.dirty:
         item.write()
      self.common.dirty = set()
      trace('done with flush')

class _KeyStorage(_Storage):
   
   def __init__(self, file = None, pos = None, common = None, key = None,
                refCount = None
                ):
      if key:
         assert refCount is not None
         self.key = key
         self.refCount = refCount
         size = 4 + len(key)
         common.dirty.add(self)
      else:
         assert pos, 'you must have either a key or a pos'
         size = None
      
      _Storage.__init__(self, file, pos, common, size)
      
      # read the key if appropriate
      if pos and not key:
         raw = self.read()
         self.key = raw[4:]
         self.refCount = struct.unpack('I', raw[:4])[0]
      
      # add it to the cache
      trace('adding key %s to %s' % (self.key, id(self.common)))
      self.common.keyStorageByKey[self.key] = self
         
   def releaseReference(self):
      self.refCount -= 1
      if not self.refCount:
         trace('removed last reference to key %s from %s' %
                (self.key, id(self.common))
               )
         del self.common.keyStorageByKey[self.key]
         if self.isAllocated():
            trace(self.common.keyStorageByPos)
            del self.common.keyStorageByPos[self.pos]
         self.markDeleted()
      else:
         self.common.dirty.add(self)
   
   def write(self):
      trace('writing storage for %s' % self.key)
      _Storage.write(self, struct.pack('I', self.refCount) + self.key)
   
   def _handlePosAssigned(self, pos):
      self.common.keyStorageByPos[pos] = self

class FullKeyVal(KeyVal, Stateful):

   def __init__(self, storage, key, val, data, state, keyStorage):
      KeyVal.__init__(self, key, val)
      self._storage = storage
      self._keyStorage = keyStorage
      self.flat = data
      self.state = state

   def _getVal(self):
      if self.state == self.STUB:
         self.flat = self._storage.read()
         self._val = Unpickler(StringIO(self.flat)).load()
         self.state = self.CLEAN
         self.proxied = False
      return self._val

   def deepStore(self):
      if self.state == self.DIRTY:
         self._storage.write(self.flat)
         self.state = self.CLEAN
   
   def _getNodeIter(self):
      yield self
   
   def _markDeleted(self):
      self._storage.markDeleted()
      self._keyStorage.releaseReference()

class FullNode(BTreeNode, Stateful):
   """
      Stored representation:
      struct {
         pointer child;
         pointer childMaxKey;
      }[MAX_SLOTS]
   """

   @staticmethod
   def getStoredSize():
      return FullNode.MAX_SLOTS * 8

   def __init__(self, type = None, maxKey = None, slots = None, 
                state = Stateful.DIRTY, 
                storage = None):
      BTreeNode.__init__(self, type, maxKey, slots)
      self.__state = state
      self._storage = storage

   def _createNode(self, type, maxKey, slots):
      storage = self._storage.alloc(self.getStoredSize())
      return self.__class__(type, maxKey, slots, storage = storage)

   def _createKeyVal(self, key, val):
      return _createKeyVal(self._storage, key, val,
                           keyStorage = self._storage.getKeyStorage(key)
                           )

   def __retrieve(self):
      # read the block
      data = self._storage.read()
      raw = array.array('I', data)
      slots = []
      self._setType(self.BRANCH)
      for i in range(0, self.MAX_SLOTS * 2, 2):
         if raw[i]:

            # see if we need to create a FullKeyVal
            if raw[i] & 0x80000000:
               childPos = raw[i] & 0x7FFFFFFF
               isKeyVal = True
            else:
               childPos = raw[i]
               isKeyVal = False

            childStorage = self._storage.create(childPos)
            keyStorage = self._storage.getKey(raw[i + 1])

            if isKeyVal:
               self._setType(self.LEAF)
               slots.append(_createKeyVal(childStorage, keyStorage.key, None,
                                          Stateful.STUB,
                                          keyStorage
                                          )
                            )
            else:
               slots.append(self.makeProxy(childStorage, keyStorage.key))
      self._setSlots(slots)

      # for the root node, maxKey is set to None.  All other nodes are created 
      # as proxies.
      if self._maxKey is None:
         self._maxKey = slots[-1]._maxKey
      else:
         # since we're a proxy, the max key should already be correct
         assert self._maxKey == slots[-1]._maxKey, \
            "at %d: parent max key %s != computed max key %s" % \
            (self._storage.pos, self._maxKey, slots[-1]._maxKey)

   def _insureReadable(self):
      if self.__state == self.STUB:
         self.__retrieve()
         self.__state = self.CLEAN
   
   def _markModified(self):
      self.__state = self.DIRTY
   
   def _markDeleted(self):
      self._storage.markDeleted()

   def __store(self):
      raw = array.array('I')
      pad = self.MAX_SLOTS
      for slot in self._getSlots():

         # flag child nodes with a high-bit
         childPos = slot._storage.pos
         if isinstance(slot, FullKeyVal):
            childPos |= 0x80000000

         # get the storage position of the slot and the key
         raw.append(childPos)
         maxKeyPos = self._storage.addKey(slot._maxKey).pos
         raw.append(maxKeyPos)

         pad -= 1

      # pad the rest of the slots with zeroes
      while pad:
         raw.append(0)
         raw.append(0)
         pad -= 1

      # and finally, persist it
      self._storage.write(raw.tostring())

      self.__state = self.CLEAN

   def deepStore(self):
      """
         Store the object and all dirty children.  This is a temporary method,
         what I want is to store the tree asynchronously from a background
         thread.
      """
      if self.__state == self.DIRTY:
         self.__store()
      for slot in self._getSlots():
         slot.deepStore()

   @staticmethod
   def makeProxy(storage, key = None):
      return FullNode(maxKey = key, state = FullNode.STUB, storage = storage)

   def _getNodeIter(self):
      self._insureReadable()
      yield self
      
      for slot in self._getSlots():
         for subSlot in slot._getNodeIter():
            yield subSlot
   
   def dump(self, indenter):
      if self._storage.isAllocated():
         print '%s at %d:' % (indenter, self._storage.pos)
      BTreeNode.dump(self, indenter)

class FullBTreeDB(MemBTreeDB):

   MAGIC = 'SpDB\0\0\0\001'

   def _createRoot(self):
      return self.__rootPos and \
         FullNode.makeProxy(self.__rootStorage.create(self.__rootPos)) or None

   def _setRoot(self, root):
      self.__rootDirty = True
      MemBTreeDB._setRoot(self, root)

   def _createNode(self, type, maxKey, slots):
      """
         Factory method to create BTree Nodes.
      """
      return FullNode(type, maxKey, slots, 
                      storage = 
                         self.__rootStorage.alloc(FullNode.getStoredSize())
                      )

   def _createKeyVal(self, key, val):
      return _createKeyVal(self.__rootStorage, key, val, 
                           keyStorage = self.__rootStorage.getKeyStorage(key)
                           )

   def __init__(self, env, path, name, allowDups):
      fullName = os.path.join(path, name)
      if os.path.exists(fullName):
         self.__fp = FileWrapper(open(fullName, 'rb+'), False)

         # check the magic number
         assert self.__fp.read(8) == self.MAGIC

         # get the root record
         self.__rootPos = struct.unpack('I', self.__fp.read(4))[0]

         self.__heap = HeapFile(self.__fp, 12)
      else:
         self.__fp = FileWrapper(open(fullName, 'wb+'), True)
         self.__fp.write(self.MAGIC)

         # write the pointer to the root record - 0 indicates that there is no
         # root record yet
         self.__fp.write('\0' * 4)
         self.__rootPos = 0

         self.__heap = HeapFile(self.__fp, 12, True)
         self.setWritable(False)
      
      # keeps track whether the pointer to the root node has been modified
      self.__rootDirty = False

      # create root storage - there's trickery and deceit here: we are
      # pointing to a location that is outside of the file heap that is
      # referenced by the storage object.
      # the heap will be used to create storage objects for everything else.
      # The size and position will be used to allow us to conveniently store
      # the root pointer when we change it.
      self.__rootStorage = _Storage(self.__heap, pos = 8, size = 4)

      # we do this after our initializations because we need rootPos to create
      # the root node
      MemBTreeDB.__init__(self, env, name, allowDups)
   
   def setWritable(self, writable):
      self.__fp.writable = writable
   
   def deepStore(self):
      root = self._getRoot()
      
      # allow writes
      self.setWritable(True)
      
      # rewrite the root pointer if it has changed
      if self.__rootDirty:
         if root is None:
            self.__rootStorage.write(struct.pack('I', 0))
         else:
            self.__rootStorage.write(struct.pack('I', root._storage.pos))
   
      # write every dirty node in the tree   
      if root:
         root.deepStore()
      
      # ... and write out everything in the storage
      self.__rootStorage.flush()
   
      self.setWritable(False)
   
   def dumpHeap(self):
      print 'block table'
      print '==========='
      self.__heap.dump()
      print '-----------'

   def checkIntegrity(self):
      """
         Does a full integrity check of the dataabase.
      """
      self.__heap.checkIntegrity()
      records = {}
      
      # add all of the records in the heap to our dictionary
      for rec in self.__heap.getBlockIter():
         assert not records.has_key(rec.pos)
         if not rec.free:
            records[rec.pos + rec.SIZE] = rec.usedSize
   
      # now go through and match up to the nodes
      root = self._getRoot()
      if root:
         for node in root._getNodeIter():
            storage = node._storage
            if storage.isAllocated():
               assert records.has_key(storage.pos) and \
                  (storage.size is None or records[storage.pos] == storage.size)
               del records[storage.pos]
      
      # note - the "root storage" exists outside of the managed heap file, so 
      # we don't have to check that.
      
      # check the deleted storage areas
      for storage in self.__rootStorage.common.deleted:
         assert records.has_key(storage.pos) and \
            storage.size == records[storage.pos]
         del records[storage.pos]
      
      # check the keys
      for key, storage in self.__rootStorage.common.keyStorageByKey.items():
         if storage.isAllocated():
            assert records.has_key(storage.pos) and \
                  records[storage.pos] == len(key) + 4
            del records[storage.pos]

      # make sure there are no allocated records that are not accounted for. 
      assert not records

class FullEnv(MemEnv):

   def __init__(self, path):
      MemEnv.__init__(self, path)
      
      # create the ".full" marker file
      file(os.path.join(path, '.full'), 'w')

   def _loadState(self):
      # XXX load the database caches
      self._tables = {}

   def _createBTreeDB(self, name, allowDups):
      return FullBTreeDB(self, self._path, name, allowDups)
   
   def _checkpointTables(self):
      for table in self._tables.values():
         table.deepStore()
   
   def checkIntegrity(self):
      """
         Does a full internal integrity check of the database.
      """
      for table in self._tables.values():
         table.checkIntegrity()
