#==============================================================================
#
"""
    Simple memory database implementation.
"""
#
#  Copyright (C) 2007 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import threading, os, sys, traceback, fcntl
from pickle import Pickler, Unpickler
from odbi import *
from weakref import WeakKeyDictionary, ref

import struct
from cStringIO import StringIO

class _Action:
    """
        An action (Command pattern implementation)
    """

    def doIt(self, env):
        "Apply the action to the database environment"
        raise NotImplementedError()

    def undoIt(self, env):
        "Undo the action on the environment"
        raise NotImplementedError()

    def dump(self, out, indent = ''):
        out.write('%s%s {\n' % 
                   (indent, self.__class__.__name__))
        for key, val in self.__dict__.items():
            out.write('%s  %s = %s\n' % (indent, key, repr(val)))
        out.write('%s}\n' % indent)

class _CreateBTreeDBAction(_Action):
    "Action to create a BTree table"

    def __init__(self, name, allowDups):
        self.name = name
        self.allowDups = allowDups

    def doIt(self, env):
        env._tables[self.name] = env._createBTreeDB(self.name, self.allowDups)

    def undoIt(self, env):
        try:
            del env._tables[self.name]
        except KeyError:
            pass

class _InsertAction(_Action):

    def __init__(self, name, key, val, oldVal, gotOldVal):
        self.name = name
        self.key = key
        self.val = val
        self.oldVal = oldVal
        self.gotOldVal = gotOldVal

    def __setstate__(self, d):
        self.__dict__ = d

    def doIt(self, env):
        env._tables[self.name]._insert(self.key, self.val)
    
    def undoIt(self, env):
        db = env._tables[self.name]
        if self.gotOldVal:
            db._insert(self.key, self.oldVal)
        else:
            db._delete(self.key)

class _DeleteAction(_Action):
    def __init__(self, name, key, oldVal):
        self.name = name
        self.key = key
        self.oldVal = oldVal

    def doIt(self, env):
        env._tables[self.name]._delete(self.key)

    def undoIt(self, env):
        env._tables[self.name]._insert(self.key, self.oldVal)

class _CreateHashDBAction(_Action):
    "Action to create a hash db"

    def __init__(self, name, allowDups):
        self.name = name
        self.allowDups = allowDups

    def doIt(self, env):
        env._tables[self.name] = MemHashDB(env, self.name)

    def undoIt(self, env):
        try:
            del env._tables[self.name]
        except KeyError:
            pass

class DoubleLockError(Exception):
    """
        Raised when an attempt is made to lock an environment that is already
        locked.  This should only happen when you create more than one
        transaction on the same environment.
    """
    pass

class SafeLock:
    """
        Lock implementation that raises an exception if it is locked more than
        once from the same thread.
    """

    def __init__(self):
        self.__impl = threading.RLock()
        self.__counter = 0
        self.__traceback = None

    def acquire(self):
        self.__impl.acquire()
        if self.__counter:
            self.__impl.release()
            raise DoubleLockError('Double locked database environment.  '
                                   'Previous lock acquired here:' +
                                   ''.join(self.__traceback)
                                  )
        self.__traceback = traceback.format_stack()
        self.__counter += 1

    def release(self):
        assert self.__counter > 0
        self.__counter -= 1
        self.__traceback = ['NOT LOCKED!']
        self.__impl.release()

class MemEnv(Env):

    def __loadTxn(self):
        data = self.__log.read(4)
        if not data:
            return None
        size = struct.unpack('I', data)[0]
        data = self.__log.read(size)
        return Unpickler(StringIO(data)).load()

    def __dumpTxn(self, txn):
        dst = StringIO()
        pickler = Pickler(dst)
        pickler.dump(txn)
        data = dst.getvalue()
        self.__log.write(struct.pack('I', len(data)))
        self.__log.write(data)

    def __readLog(self):
        """
            Read all outstanding transactions in the logfile.
        """
        while True:
            txn = self.__loadTxn()
            if not txn:
                break
            txn._apply(self)
#                    try:
#                        log.load()._apply(self)
#                    except EOFError, ex:
#                        break

    def _loadState(self):
        # load the big state file, if it exists
        stateFileName = os.path.join(self._path, 'State')
        if os.path.exists(stateFileName):
            self._tables = Unpickler(file(stateFileName, 'rb')).load()

            # restore the reference to the environment
            for table in self._tables.values():
                table._env = self

        else:
            # create an empty database
            self._tables = {}

    def __init__(self, path):

        self._path = path

        # database-wide lock
        self._lk = SafeLock()

        # transaction used for all cursor operations
        self.__tmpTxn = None

        # make sure the directory exists
        if not os.path.exists(path):
            os.mkdir(path)
            file(os.path.join(path, '.mem'), 'w')

        # make sure that a lock file exists (XXX race condition)
        lockFile = os.path.join(self._path, 'lock')
        if not os.path.exists(lockFile):
            self.__lock = file(lockFile, 'w')
        else:
            self.__lock = file(lockFile, 'r+')

        # get the lock, but don't try to read from the logfile yet
        self._acquireLock(update = False)

        try:
            # load the current state
            self._loadState()

            # if there's a logfile, open it and read it
            lastLog = self.__getLastLog()
            if lastLog:
                lastLog = os.path.join(path, lastLog)
                self.__log = file(lastLog, 'rb+')
                self.__readLog()
            else:
                self.__openNewLog()
        finally:
            self._releaseLock()

    def __getLastLog(self):
        # find the last logfile - MUST HAVE THE DIRECTORY LOCK
        lastLog = None
        files = os.listdir(self._path)
        files.sort()
        for filename in files:
            if filename.startswith('log'):
                lastLog = filename
        return lastLog

    def __openNewLog(self):
        # open a new logfile - MUST HAVE THE DIRECTORY LOCK
        lastLog = self.__getLastLog()
        if lastLog:
            counter = int(lastLog.split('.')[-1]) + 1
        else:
            counter = 1
        self.__log = \
            file(os.path.join(self._path, 'log.%09d' % counter), 'wb+')

    def _createBTreeDB(self, name, allowDups):
        return MemBTreeDB(self, name, allowDups)

    def getBTreeDB(self, name, allowDups = False):

        # first check the cache
        if self._tables.has_key(name):
            db = self._tables[name]
            assert isinstance(db, MemBTreeDB)
            return db

        # create it transactionally
        txn = MemTxn(self)
        try:
            txn._add(_CreateBTreeDBAction(name, allowDups))
            self._tables[name] = db = self._createBTreeDB(name, allowDups)
            txn.commit()
            txn = None
        finally:
            if txn: txn.abort()

        return db

    def _createHashDB(self, name, allowDups):
        return MemHashDB(self, name)

    def getHashDB(self, name, allowDups = False):

        # check the cache
        if self._tables.has_key(name):
            db = self._tables[name]
            assert isinstance(db, MemHashDB)
            return db

        # create it transactionally
        txn = MemTxn(self)
        try:
            txn._add(_CreateHashDBAction(name, allowDups))
            self._tables[name] = db = self._createHashDB(name, allowDups)
            txn.commit()
            txn = None
        finally:
            if txn: txn.abort()

        return db

    def getTxn(self):
        return MemTxn(self)

    def checkpoint(self):
        self._acquireLock()
        try:
            self._checkpointTables()
            self.__openNewLog()
        finally:
            self._releaseLock()
    
    def close(self):
        self._acquireLock()
        try:
            for table in self._tables.values():
                table.close()
            
            self.__log.close()
            self.__log = None
        finally:
            self._releaseLock()
    
    def _checkpointTables(self):
        stateFileName = os.path.join(self._path, 'State')
        stateFile = open(stateFileName, 'wb')
        Pickler(stateFile).dump(self._tables)

    def startDeadlockThread(self, deadlockCheckInterval, noticeCallback):
        pass

    def _acquireLock(self, update = True):
        # use both a mutex and a file system lock - the first should perform
        # better for locking between threads (and should also give us a nice
        # sane error if we double-lock), the second is essential for locking
        # between processes
        self._lk.acquire()
        fcntl.lockf(self.__lock.fileno(), fcntl.LOCK_EX)

        # read any transactions that we might have missed
        if update: self.__readLog()

    def _releaseLock(self):
        fcntl.lockf(self.__lock.fileno(), fcntl.LOCK_UN)
        self._lk.release()

    def _write(self, txn):
        self.__dumpTxn(txn)
        #self.__log.dump(txn)
        self.__log.flush()

    def _getTmpTxn(self):
        txn = self.__tmpTxn and self.__tmpTxn()
        if not txn:
            txn = MemTxn(self)
            self.__tmpTxn = ref(txn)
        return txn
    
    def _removeTable(self, name):
        del self._tables[name]

class MemCursor(Cursor):
    """
        Memory-oriented implementation of a database cursor.
    """

    def __init__(self, db, state, txn):
        self.__db = db
        self.__state = state
        self.__dir = 1
        self.__txn = txn
        txn._addCursor(self)

    def __del__(self):
        self.close()

    def close(self):
        if self.__db:
            self.__txn._removeCursor(self)
            self.__db = self.__state = self.__dir = self.__txn = None

    def setToFirst(self):
        assert self.__db
        self.__state = self.__db._first()
        self.__dir = 1

    def setToLast(self):
        assert self.__db
        self.__state = self.__db._last()
        self.__dir = -1

    def setToKey(self, key, exact = False):
        assert self.__db
        self.__state = self.__db._find(key, exact)
        self.__dir = 1

    def next(self):
        assert self.__db
        # since the cursor is actually a path to the next available node, and
        # we want it to behave like a path to the point _before_ the next
        # available node, if we are currently moving in the opposite
        # direction, we have to do "next" twice
        if self.__dir == -1 and self.__state and self.__state[0] != 0: 
            self.__db._next(self.__state)
        self.__dir = 1
        self.__state, key, val = self.__db._next(self.__state)
        return key, val

    def prev(self):
        assert self.__db
        # reverse direction if necessary
        if self.__dir == 1 and self.__state and self.__state[0] != -1: 
            self.__db._prev(self.__state)
        self.__dir = -1
        self.__state, key, val = self.__db._prev(self.__state)
        return key, val

    def __iter__(self):
        return self

class MemTxn(Txn):
    """
        Transaction implementation for memory databases.
    """

    def __init__(self, env):
        self.__env = None
        env._acquireLock()
        self.__env = env
        self.__actions = []
        self.__cursors = WeakKeyDictionary()

    def __del__(self):
        if self.__env:
            self.abort()

    def _addCursor(self, cursor):
        self.__cursors[cursor] = True

    def _removeCursor(self, cursor):
        if self.__cursors.has_key(cursor):
            del self.__cursors[cursor]

    def _apply(self, env):
        # apply all actions to the database
        for action in self.__actions:
            action.doIt(env)
        self.__actions = None

    def _add(self, action):
        self.__actions.append(action)

    def __setstate__(self, d):
        self.__dict__ = d
        self.__env = None
        self.__cursors = WeakKeyDictionary()

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_MemTxn__env']
        del d['_MemTxn__cursors']
        return d

    def __commonClose(self):
        "Clean-up function to be called both by commit and abort"

        try:
            if self.__cursors:
                # close all of the cursors, we have to use "keys()" to collect
                # them instead of straight out iteration because Cursor.close()
                # removes the cursor from the __cursors dictionary and we would
                # invalidate the iterator.
                for cur in self.__cursors.keys():
                    cur.close()
        finally:
            # if there's a screw-up with closing the cursor, we still need to
            # release the lock and release the environment
            self.__actions = None
            if self.__env: self.__env._releaseLock()

            # disconnect so we don't try to do this again
            self.__env = None

    def commit(self):
        # if there are any actions, commit the transaction to the logfile
        try:
            if self.__actions:
                self.__env._write(self)
        finally:
            self.__commonClose()

    def abort(self):
        # unwind all of the actions that we performed
        try:
            if self.__actions:
                self.__actions.reverse()
                for action in self.__actions:
                    action.undoIt(self.__env)
        finally:
            self.__commonClose()

    def dump(self, out, indent = ''):
        out.write(indent + 'Txn {\n')
        for act in self.__actions:
            act.dump(out, indent + '  ')
        out.write(indent + '}\n')

class KeyVal(object):

    # define two "virtual" methods and use lambda to associate them with a 
    # property (otherwise they don't get virtualized)
    def _getVal(self): return self._val
    def _setVal(self, val): self._val = val
    val = property(lambda self: self._getVal(), 
                   lambda self, val: self._setVal(val)
                   )

    def __init__(self, key, val):
        self.key = key
        self._val = val
        self._maxKey = key

    def dump(self, indenter):
        print indenter, 'key = %s, val = %s' % \
            (repr(self.key), repr(self.val))
    
    def _markDeleted(self):
        "Allows derived classes to mark the node for deletion"
        pass

class BTreeNode:
    """
        BTree tree node.
    """

    # node types - a node can be a leaf or a branch
    LEAF = 0
    BRANCH = 1

    # max and minimum children for each node
    MAX_SLOTS = 3
    MIN_SLOTS = 2

    def __init__(self, type, maxKey, slots):
        self.__type = type
        self.__slots = slots
        self._maxKey = maxKey

    def _setSlots(self, slots):
        self.__slots = slots

    def _getSlots(self):
        return self.__slots

    def _setType(self, type):
        self.__type = type

    def _createNode(self, type, maxKey, slots):
        return self.__class__(type, maxKey, slots)

    def _createKeyVal(self, key, val):
        return KeyVal(key, val)

    def find(self, key, path = None, exact = True):
        """
            Finds the value for the key, optionally filling in the path to the
            node.  Returns a tuple of the value (or None if the key was not
            found) and a boolean indicated whether the key was found

            parms:
                key::
                    [string] the key
                path::
                    [list<BTreeNode> or None] if specified, this is a list
                    that we will fill in with the nodes forming the path.
                exact::
                    [boolean] if true, will only find an exact match.
                    Otherwise, finds the first key that is greater than or
                    equal to the specified key.
        """
        self._insureReadable()

        # make sure that the key might possibly be in this node.
        # this should only happen for the root node, but we have to deal
        # with it
        if key > self._maxKey:
            path = [-1]
            return None, False

        # if the caller has requested a path, add this node to it
        if path is not None:
            path.append(self)

        if self.__type == self.LEAF:
            updatedPath = False
            for slot in self.__slots:

                # if the caller has requested a path and the next slot either
                # matches or exceeds the key, store the key/value node in the
                # path
                if slot.key >= key and path is not None and not updatedPath:
                    path.append(slot)
                    updatedPath = True

                # if we've found a match, return it
                if slot.key == key:
                    return slot.val, True

                # if we've found a greater key, either this is a match (if
                # we're not looking for an exact match) or the key is
                # undefined
                elif slot.key > key:
                    if exact:
                        return None, False
                    else:
                        return slot.val, True
            else:
                # for us to end up here, the key we're looking for has to be
                # between the last key in the slot and the _maxKey instance
                # variable, and that, of course, should never happen
                assert False, \
                    'broken invarient in leaf node, maxKey = %s, slot.key ' \
                    '= %s, key = %s' % (self._maxKey, slot.key, key)

        else:
            for slot in self.__slots:
                if key <= slot._maxKey:
                    return slot.find(key, path)
            else:
                # key is after the last key - we're at the end.
                if path is not None:
                    path[:] = [-1]
                if not exact:
                    return None, True

        # if we haven't added anything to the path, clear the path
        if path is not None and path[-1] is self:
            path[:] = [-1]

        return None, False

    def _insureReadable(self):
        """
            Called whenever we are about to perform a read operation.  Allows 
            a derived class to implement lazy-load.
        """
        pass
    
    def _markModified(self):
        """
            Called whenever the node has been modified.  Allows a derived 
            class to track if the node is dirty.
        """
        pass
    
    def _markDeleted(self):
        """
            Called whenever the node is deleted.  Like @_markModified(), this 
            is in the interests of allowing derived classes to track deletion.
        """
        pass

    def insert(self, key, val, allowDups):
        self._insureReadable()
        if self.__type == self.LEAF:
            for slot, i in zip(self.__slots, xrange(len(self.__slots))):
                if not allowDups and key == slot.key:
                    self.__slots[i] = self._createKeyVal(key, val)
                    break
                elif key < slot.key:
                    self.__slots.insert(i, self._createKeyVal(key, val))
                    break
            else:
                self.__slots.append(self._createKeyVal(key, val))
                self._maxKey = key

            self._markModified()
            maxSlots = self.MAX_SLOTS

        else:
            # delegate to children
            for child, i in zip(self.__slots, xrange(len(self.__slots))):
                if key <= child._maxKey:
                    newNode = child.insert(key, val, allowDups)
                    break
            else:
                # insert into the last node, fix the last key
                newNode = child.insert(key, val, allowDups)
                self._maxKey = key
                self._markModified()

            # if we did a split, we have to insert the new node before the
            # current node
            if newNode:
                self.__slots.insert(i, newNode)
                self._markModified()

            maxSlots = self.MAX_SLOTS

        # see if we need to split
        if len(self.__slots) > maxSlots:

            # split the slot array
            mid = len(self.__slots) / 2
            newSlots = self.__slots[:mid]
            self.__slots = self.__slots[mid:]

            # create a new peer node - new nodes must preceed
            # the node that they were split off of
            newLeaf = \
                self._createNode(self.__type, newSlots[-1]._maxKey, newSlots)

            return newLeaf
            
            # we shouldn't need to "markModified()" - if we need to split, 
            # we've already been modified

        else:
            return None

    def redistribute(self, sibling):
        """
            If the sibling has more than the minimum number of slots, steal
            some of his and return true.  Otherwise return false.
        """

        self._insureReadable()

        # if the sibling has more than the minimum number of slots, take his
        # surplus
        if len(sibling.__slots) > self.MIN_SLOTS:
            if self._maxKey > sibling._maxKey:
                # sibling preceeds us - construct from the top half of his
                # nodes then ours
                self.__slots = \
                    sibling.__slots[self.MIN_SLOTS:] + self.__slots
                sibling.__slots = sibling.__slots[:self.MIN_SLOTS]
                sibling._maxKey = sibling.__slots[-1]._maxKey
            else:
                # we preceed sibling - construct from our nodes and then  the
                # bottom half of his
                self.__slots = \
                    self.__slots + sibling.__slots[:self.MIN_SLOTS - 1]
                self._maxKey = self.__slots[-1]._maxKey
                sibling.__slots = sibling.__slots[self.MIN_SLOTS - 1:]
            
            self._markModified()
            sibling._markModified()

            return True
        else:
            return False

    def merge(self, sibling):

        self._insureReadable()

        # this function assumes that the sibling has only the minimum number
        # of nodes
        assert len(sibling.__slots) == self.MIN_SLOTS and \
               len(self.__slots) < self.MIN_SLOTS

        if self._maxKey > sibling._maxKey:
            self.__slots = sibling.__slots + self.__slots
        else:
            self.__slots = self.__slots + sibling.__slots
            self._maxKey = sibling._maxKey
        
        self._markModified()
        sibling._markDeleted()

        return self

    def remove(self, key):
        """
            Removes the key from the node and its children.

            If the operation leaves the node empty, returns False - the parent
            must remove the node in this case.

            Returns the value of the deleted node (*None* if it wasn't found)
            and the number of children remaining.
        """
        self._insureReadable()

        if self.__type == self.LEAF:
            for slot in self.__slots:
                if slot.key == key:
                    self.__slots.remove(slot)
                    slot._markDeleted()
                    self._markModified()

                    # if the key was the maximum key, fix the maximum key
                    if key == self._maxKey and self.__slots:
                        self._maxKey = self.__slots[-1].key
                        assert self._maxKey == self.__slots[-1]._maxKey

                    return slot.val, len(self.__slots)
                elif key < slot.key:
                    raise KeyError(key)
            else:
                # key not found (this should only occur if the leaf is the
                # root)
                raise KeyError(key)
        else:
            i = 0
            for slot in self.__slots:
                if key <= slot._maxKey:
                    slotMaxKey = self._maxKey
                    val, children = slot.remove(key)
                    
                    # if a slot's max key changed, the parent  becomes dirty
                    # (this is more tightly coupled to the implementation of 
                    # the derived class than I'm comfortable with, but the 
                    # only alternative I can see is to provide a different 
                    # flavor of "markModified()" for this case, and I like 
                    # that even less)
                    if slot._maxKey != slotMaxKey:
                        self._markModified()
                    
                    # fix our maximum key
                    newMax = self.__slots[-1]._maxKey
                    if self._maxKey != newMax:
                        self._maxKey = newMax
                        self._markModified()
                        
                    if not children:
                        # the child now has no children ... remove it
                        # XXX this should never happen except for the root
                        # node.  Remove me
                        self.__slots.remove(slot)

                        # adjust the max key if necessary
                        if self.__slots and key == self._maxKey:
                            self._maxKey = self.slots[-1]._maxKey
                        
                        self._markModified()

                    elif children < self.MIN_SLOTS and len(self.__slots) > 1:
                        
                        # rebalance the nodes

                        # if this is not the first slot, try to redistribute
                        # with the slot before it
                        if i > 0 and slot.redistribute(self.__slots[i - 1]):
                            pass

                        # if this is not the last slot, try to redistribute
                        # with the one after it
                        elif i < len(self.__slots) - 1 and \
                            slot.redistribute(self.__slots[i + 1]):
                            pass

                        # if that fails, merge with either of them - we rely 
                        # on @merge() to mark the sibling as deleted
                        elif i > 0:
                            slot.merge(self.__slots[i - 1])
                            del self.__slots[i - 1]
                        else:
                            slot.merge(self.__slots[i + 1])
                            del self.__slots[i + 1]
                        
                        self._markModified()

                    return val, len(self.__slots)

                i += 1
            else:
                # key is greater than all in the set (should only occur for
                # the root node)
                raise KeyError(key)

        return None, 0
    
    def collapse(self):
        """
            If the node is not a leaf, collapses the node by marking it 
            deleted and returning its first child.  This should only be done 
            to a node with one child (as indicated by the return value of 
            @remove()).
            
            In any case, returns the new toplevel node (self or its child).
        """
        if self.__type == self.BRANCH:
            self._markDeleted()
            return self.__slots[0]
        else:
            return self

    def dump(self, indenter):
        self._insureReadable()
        print indenter, 'FullNode(%s)' % id(self), \
            self.__type == self.LEAF and 'LEAF' or 'BRANCH', \
            'maxKey = %s' % self._maxKey
        for child in self.__slots:
            child.dump(indenter + ' ')

    def getPathToFirst(self):
        self._insureReadable()
        if self.__type == self.LEAF:
            return [self, self.__slots[0]]
        else:
            return [self] + self.__slots[0].getPathToFirst()

    def getPathToLast(self):
        self._insureReadable()
        if self.__type == self.LEAF:
            return [self, self.__slots[-1]]
        else:
            return [self] + self.__slots[-1].getPathToLast()

    class ForwardDirection:
        increment = 1
        startIndex = 0
        endMarker = -1
        def outOfRange(self, index, slots):
            return index >= len(slots)
    forwardDirection = ForwardDirection()

    class ReverseDirection:
        increment = -1
        startIndex = -1
        endMarker = 0
        def outOfRange(self, index, slots):
            return index < 0
    reverseDirection = ReverseDirection()

    @staticmethod
    def next(path):
        return BTreeNode.__next(path, BTreeNode.forwardDirection)

    @staticmethod
    def prev(path):
        return BTreeNode.__next(path, BTreeNode.reverseDirection)

    @staticmethod
    def __next(path, direction):
        assert len(path) >= 2
        # store the current node - it's what we're returning
        current = path[-1]

        # consider the bottom two levels - find the currently referenced child
        # of the current parent
        parent = path[-2]
        nextIndex = parent.__slots.index(path[-1]) + direction.increment

        # if the next index is out of range, find the uncle through the
        # grandparent
        if direction.outOfRange(nextIndex, parent.__slots): 

            # if there _is_ no grandparent, we're done
            if len(path) == 2:
                path[:] = [direction.endMarker]
                return current

            path.pop(-1)
            parent.__next(path, direction)
            if not path or isinstance(path[0], int):
                return current
            uncle = path[-1]

            # make our "first cousin" the next node
            uncle._insureReadable()
            path.append(uncle.__slots[direction.startIndex])
        else:
            path[-1] = parent.__slots[nextIndex]
        
        return current

class MemKeyDB(KeyDB):
    """
        Abstract implementation of KeyDB which defines all of the standard
        functions in terms of the hidden implementation functions.
    """

    def __init__(self, env, name):
        self._env = env
        self._name = name

    def __getstate__(self):
        dict = self.__dict__.copy()
        del dict['_env']
        return dict

    def set(self, txn, key, val):
        # store the current value
        oldVal, gotVal = self._lookUp(key)

        # insert the new value
        self._insert(key, val)

        # store the action
        act = _InsertAction(self._name, key, val, oldVal, gotVal)
        txn._add(act)

    def __getitem__(self, key):
        txn = self._env._getTmpTxn()
        val, gotVal = self._lookUp(key)
        del txn
        if not gotVal:
            raise KeyError(key)
        else:
            return val

    def get(self, key, default = None, txn = None):
        if not txn:
            txn = myTxn = self._env._getTmpTxn()
        val, gotVal = self._lookUp(key)
        if not gotVal:
            return default
        else:
            return val

    def has_key(self, key):
        txn = self._env._getTmpTxn()
        val, gotVal = self._lookUp(key)
        return gotVal

    def delete(self, txn, key):
        val = self._delete(key)
        # add it to the transaction
        txn._add(_DeleteAction(self._name, key, val))

    def __delitem__(self, key):
        txn = MemTxn(self._env)
        self.delete(txn, key)
        txn.commit()

    def __setitem__(self, key, val):
        txn = MemTxn(self._env)
        try:
            self.set(txn, key, val)
            txn.commit()
            txn = None
        finally:
            if txn: txn.abort()

    def getTxn(self):
        return self._env.getTxn()
    
    def close(self):
        self._env._removeTable(self._name)
        self._env = None

class MemBTreeDB(MemKeyDB):
    """
        In-memory b-tree database implementation.
    """

    def __init__(self, env, name, allowDups):
        MemKeyDB.__init__(self, env, name)
        self.__root = self._createRoot()
        self.__allowDups = allowDups
        self._name = name

    def _createRoot(self):
        return None

    def _setRoot(self, root):
        self.__root = root
    
    def _getRoot(self):
        return self.__root

    def _createNode(self, type, key, slots):
        return BTreeNode(type, key, slots)

    def _createKeyVal(self, key, val):
        return KeyVal(key, val)

    def _write(self):
        "Write the entire state"

    def _lookUp(self, key):
        """
            Looks up the key, returns a tuple of the value (or None if not
            found) and a boolean set to true if the key was found.
        """
        if self.__root:
            return self.__root.find(key)
        else:
            return None, False

    def _insert(self, key, val):
        if self.__root:
            newNode = self.__root.insert(key, val, self.__allowDups)
            if newNode:
                self._setRoot(
                    self._createNode(BTreeNode.BRANCH, self.__root._maxKey,
                                     [newNode, self.__root]
                                     )
                )
        else:
            self._setRoot(
                self._createNode(BTreeNode.LEAF, key, 
                                 [self._createKeyVal(key, val)]
                                 )
            )

    def _delete(self, key):
        if not self.__root:
            raise KeyError(key)
        val, childCount = self.__root.remove(key)
        
        # if the root node is now empty, discard it
        if not childCount:
            self.__root._markDeleted()
            self._setRoot(None)

        # if there is just one child, set the root to the child
        if childCount == 1:
            root = self.__root.collapse()
            if root is not self.__root:
                self._setRoot(root)

        return val

    def cursor(self, txn = None):
        if not self.__root:
            path = None
        else:
            path = [0]

        # make sure that we do this in the scope of a transaction
        if not txn:
            txn = self._env._getTmpTxn()

        return MemCursor(self, path, txn)

    def _getTmpTxn(self):
        return self._env._getTmpTxn()

    def __first(self):
        return self.__root.getPathToFirst()

    def __last(self):
        return self.__root.getPathToLast()

    def _next(self, path):
        if not path or path[0] == -1:
            raise StopIteration()
        else:
            if path[0] == 0: path[:] = self.__first()
            node = BTreeNode.next(path)
            return path, node.key, node.val

    def _prev(self, path):
        if not path or path[0] == 0:
            raise StopIteration()
        else:
            if path[0] == -1: path[:] = self.__last()
            node = BTreeNode.prev(path)
            return path, node.key, node.val

    def _find(self, key, exact):
        path = []
        if self.__root:
            self.__root.find(key, path, exact)
        if not path and self.__root:
            if not exact and key > self.__root._maxKey:
                path = [-1]
            elif not exact:
                path = [0]

        return path

    def _first(self): return [0]
    def _last(self): return [-1]

    def dump(self):
        if self.__root:
            self.__root.dump('')
        else:
            print '<empty db>'

class MemHashDB(MemKeyDB):
    """
        Hash table implementation.
    """

    def __init__(self, env, name):
        MemKeyDB.__init__(self, env, name)
        self.__dict = {}

    def _insert(self, key, val):
        self.__dict[key] = val

    def _delete(self, key):
        del self.__dict[key]

    def _lookUp(self, key):
        if self.__dict.has_key(key):
            return self.__dict[key], True
        else:
            return None, False

    # this is badly broken - but who cares about this for a hash db anyhow?
    def _find(self, key, exact): return self.__dict.iteritems()
    def _first(self): return self.__dict.iteritems()
    def _last(self): return self.__dict.iteritems()

    def _next(self, state):
        return (state,) + state.next()

    def cursor(self, txn = None):
        if not txn:
            txn = self._env._getTmpTxn()
        return MemCursor(self, self.__dict.iteritems(), txn)

if __name__ == '__main__':

    # this stuff needs to be done in a function because if we let it get
    # cleaned up in the global destructors we get an exception that appears to
    # be related to the order of destruction of the system objects
    def testfunc():
        print 'running test suite'
        import shutil
        if os.path.exists('utdata'): shutil.rmtree('utdata')
        env = MemEnv('utdata')
        db = env.getBTreeDB('First')

        # check the empty cursor condition
        print 'checking empty cursors'
        count = 0
        for key, val in db.cursor():
            count += 1
        assert count == 0

        # put some items in the database, make sure we get them all out
        print 'testing insertions'
        elems = [
            ('hello', 'world'),
            ('first', 100),
            ('second', 200),
            ('force tree', 300),
        ]
        elems.sort()
        sortedElems = elems[:]
        for elem in elems:
            db[elem[0]] = elem[1]

        # make sure that everything comes back out in order
        print 'checking ordered traversal'
        cur = db.cursor()
        for key, val in cur:
            assert (key, val) == elems.pop(0)
        assert elems == []

        # check reverse iteration
        print 'checking reverse iteration'
        elems = sortedElems[:]
        elems.reverse()
        cur = db.cursor()
        cur.setToLast()
        results = []
        for key, val in ReverseCursor(cur):
            assert (key, val) == elems.pop(0)

        # check cursor reversibility
        print 'checking reversibility'
        cur = db.cursor()
        key, val == cur.next()
        assert (key, val) == cur.prev()
        assert (key, val) == cur.next()
        cur.close()

        print 'checking for overwrites'
        db['hello'] = 'mother'
        db['hello'] = 'world'
        cur = db.cursor()
        for key, val in sortedElems:
            assert (key, val) == cur.next()
        try:
            cur.next()
            assert 0, "more stuff than expected in the db"
        except StopIteration:
            pass
        cur.close()

        # check the "getitem" functionality
        assert env.getBTreeDB('First')['hello'] == 'world'

        print 'checking hash databases'
        hdb = env.getHashDB('Second')
        helems = [('check', 100), ('out', 300), ('this', 200)]
        for key, val in helems:
            hdb[key] = val
        results = []
        for key, val in hdb.cursor():
            results.append((key, val))
        results.sort()
        assert results == [('check', 100), ('out', 300), ('this', 200)]

        print 'checking btrees with duplicate keys'
        ddbElems = [
            ('first', 100), 
            ('second', 200), 
            ('second', 300), 
            ('third', 400)
        ]
        ddb = env.getBTreeDB('BTreeWithDups', allowDups = True)
        for key, val in ddbElems:
            ddb[key] = val
        ddbElems.sort()
        cur = ddb.cursor()
        for key, val in ddbElems:
            #k,v = cur.next()
            assert (key, val) == cur.next()
        try:
            cur.next()
            assert 0, 'too many elements'
        except StopIteration:
            pass

        # now verify persistence
        print 'verifying persistence'
        del db, env, hdb
        env = MemEnv('utdata')
        db = env.getBTreeDB('First')
        assert db['hello'] == 'world'
        elems = sortedElems[:]
        for key, val in db.cursor():
            assert (key, val) == elems.pop(0)
        assert elems == []

        print 'verifying hash persistence'
        results = []
        hdb = env.getHashDB('Second')
        for key, val in hdb.cursor():
            results.append((key, val))
        results.sort()
        assert results == [('check', 100), ('out', 300), ('this', 200)]

        print 'hash deletion'
        txn = hdb.getTxn()
        hdb.set(txn, 'delete', 'me')
        hdb.delete(txn, 'delete')
        txn.commit()
        results = []
        for key, val in hdb.cursor():
            results.append((key, val))
        results.sort()
        assert results == [('check', 100), ('out', 300), ('this', 200)]

        # check for removal
        print 'verifying removal'
        del db[sortedElems[0][0]]
        try:
            db[sortedElems[0][0]]
            assert 0, 'removal failed'
        except KeyError, ex:
            assert ex[0] == sortedElems[0][0]
        # restore it
        db[sortedElems[0][0]] = sortedElems[0][1]

        # test checkpoints
        print 'testing checkpoints'
        env.checkpoint()

        del env, db, hdb
        env = MemEnv('utdata')
        db = env.getBTreeDB('First')
        results = []
        for key, val in db.cursor():
            results.append((key, val))
        assert results == sortedElems

        hdb = env.getHashDB('Second')
        results = []
        for key, val in hdb.cursor():
            results.append((key, val))
        results.sort()
        assert results == helems

        print 'double locking'
        try:
            txn = env.getTxn()
            txn2 = env.getTxn()
            assert 0, 'double locks are not giving errors'
        except DoubleLockError:
            pass


        print 'all tests passed'
    testfunc()

# vim:set sw=4:
